#!/usr/bin/env python

"""
BaseHTTPRequestHandler classes.
"""

import Generic
from Helpers.Request import MessageBodyStream
from Helpers.Auth import UserInfo
from cgi import parse_qs, FieldStorage
import Cookie
from StringIO import StringIO

class Transaction(Generic.Transaction):

    """
    BaseHTTPRequestHandler transaction interface.
    """

    def __init__(self, trans):

        """
        Initialise the transaction using the BaseHTTPRequestHandler instance
        'trans'.
        """

        self.trans = trans

        # Other attributes of interest in instances of this class.

        self.content_type = None
        self.response_code = 200
        self.content = StringIO()
        self.headers_out = {}
        self.cookies_out = Cookie.SimpleCookie()
        self.user = None

        # Define the incoming cookies.

        self.cookies_in = Cookie.SimpleCookie(self.get_headers().get("cookie"))

        # Cached information.

        self.storage_body = None

    def commit(self):

        """
        A special method, synchronising the transaction with framework-specific
        objects.
        """

        self.trans.send_response(self.response_code)
        if self.content_type is not None:
            self.trans.send_header("Content-Type", self.format_content_type(self.content_type))

        for header, value in self.headers_out.items():
            self.trans.send_header(self.format_header_value(header), self.format_header_value(value))

        # NOTE: May not be using the appropriate method.

        for morsel in self.cookies_out.values():
            self.trans.send_header("Set-Cookie", morsel.OutputString())

        self.trans.end_headers()
        self.content.seek(0)
        self.trans.wfile.write(self.content.read())

    # Request-related methods.

    def get_request_stream(self):

        """
        A framework-specific method which returns the request stream for
        the transaction.
        """

        return MessageBodyStream(self.trans.rfile, self.get_headers())

    def get_request_method(self):

        """
        A framework-specific method which gets the request method.
        """

        return self.trans.command

    def get_headers(self):

        """
        A framework-specific method which returns all request headers as a
        dictionary-like object mapping header names to values.
        NOTE: If duplicate header names are permitted, then this interface will
        NOTE: need to change.
        """

        return self.trans.headers

    def get_header_values(self, key):

        """
        A framework-specific method which returns a list of all request header
        values associated with the given 'key'. Note that according to RFC 2616,
        'key' is treated as a case-insensitive string.
        """

        return self.convert_to_list(self.trans.headers.get(key))

    def get_content_type(self):

        """
        A framework-specific method which gets the content type specified on the
        request, along with the charset employed.
        """

        return self.parse_content_type(self.trans.headers.get("content-type"))

    def get_content_charsets(self):

        """
        Returns the character set preferences.
        """

        return self.parse_content_preferences(self.trans.headers.get("accept-charset"))

    def get_content_languages(self):

        """
        A framework-specific method which extracts language information from
        the transaction.
        """

        return self.parse_content_preferences(self.trans.headers.get("accept-language"))

    def get_path(self):

        """
        A framework-specific method which gets the entire path from the request.
        """

        return self.trans.path

    def get_path_without_query(self):

        """
        A framework-specific method which gets the entire path from the request
        minus the query string.
        """

        # Remove the query string from the end of the path.

        return self.trans.path.split("?")[0]

    def get_path_info(self):

        """
        A framework-specific method which gets the "path info" (the part of the
        URL after the resource name handling the current request) from the
        request.
        """

        return self.get_path_without_query()

    def get_query_string(self):

        """
        A framework-specific method which gets the query string from the path in
        the request.
        """

        t = self.trans.path.split("?")
        if len(t) == 1:
            return ""
        else:

            # NOTE: Overlook erroneous usage of "?" characters in the path.

            return "?".join(t[1:])

    # Higher level request-related methods.

    def get_fields_from_path(self):

        """
        A framework-specific method which extracts the form fields from the
        path specified in the transaction. The underlying framework may refuse
        to supply fields from the path if handling a POST transaction.

        Returns a dictionary mapping field names to lists of values (even if a
        single value is associated with any given field name).
        """

        return parse_qs(self.get_query_string(), keep_blank_values=1)

    def get_fields_from_body(self):

        """
        A framework-specific method which extracts the form fields from the
        message body in the transaction.

        Returns a dictionary mapping field names to lists of values (even if a
        single value is associated with any given field name).
        """

        if self.storage_body is None:
            self.storage_body = FieldStorage(fp=self.get_request_stream(), headers=self.get_headers(),
                environ={"REQUEST_METHOD" : self.get_request_method()}, keep_blank_values=1)

        # Avoid strange design issues with FieldStorage by checking the internal
        # field list directly.

        fields = {}
        if self.storage_body.list is not None:

            # Traverse the storage, finding each field value.

            for field_name in self.storage_body.keys():
                fields[field_name] = self.storage_body.getlist(field_name)
        return fields

    def get_user(self):

        """
        A framework-specific method which extracts user information from the
        transaction.

        Returns a username as a string or None if no user is defined.
        """

        if self.user is not None:
            return self.user

        auth_header = self.get_headers().get("authorization")
        if auth_header:
            return UserInfo(auth_header).username
        else:
            return None

    def get_cookies(self):

        """
        A framework-specific method which obtains cookie information from the
        request.

        Returns a dictionary mapping cookie names to cookie objects.
        """

        return self.cookies_in

    def get_cookie(self, cookie_name):

        """
        A framework-specific method which obtains cookie information from the
        request.

        Returns a cookie object for the given 'cookie_name' or None if no such
        cookie exists.
        """

        return self.cookies_in.get(cookie_name)

    # Response-related methods.

    def get_response_stream(self):

        """
        A framework-specific method which returns the response stream for
        the transaction.
        """

        # Return a stream which is later emptied into the real stream.

        return self.content

    def get_response_code(self):

        """
        Get the response code associated with the transaction. If no response
        code is defined, None is returned.
        """

        return self.response_code

    def set_response_code(self, response_code):

        """
        Set the 'response_code' using a numeric constant defined in the HTTP
        specification.
        """

        self.response_code = response_code

    def set_header_value(self, header, value):

        """
        Set the HTTP 'header' with the given 'value'.
        """

        # The header is not written out immediately due to the buffering in use.

        self.headers_out[header] = value

    def set_content_type(self, content_type):

        """
        A framework-specific method which sets the 'content_type' for the
        response.
        """

        # The content type has to be written as a header, before actual content,
        # but after the response line. This means that some kind of buffering is
        # required. Hence, we don't write the header out immediately.

        self.content_type = content_type

    def set_cookie(self, cookie):

        """
        A framework-specific method which stores the given 'cookie' object in
        the response.
        """

        # NOTE: If multiple cookies of the same name could be specified, this
        # NOTE: could need changing.

        self.cookies_out[cookie.name] = cookie.value

    def set_cookie_value(self, name, value, path=None, expires=None):

        """
        A framework-specific method which stores a cookie with the given 'name'
        and 'value' in the response.

        The optional 'path' is a string which specifies the scope of the cookie,
        and the optional 'expires' parameter is a value compatible with the
        time.time function, and indicates the expiry date/time of the cookie.
        """

        self.cookies_out[name] = value
        if path is not None:
            self.cookies_out[name]["path"] = path
        if expires is not None:
            self.cookies_out[name]["expires"] = expires

    def delete_cookie(self, cookie_name):

        """
        A framework-specific method which adds to the response a request that
        the cookie with the given 'cookie_name' be deleted/discarded by the
        client.
        """

        # Create a special cookie, given that we do not know whether the browser
        # has been sent the cookie or not.
        # NOTE: Magic discovered in Webware.

        self.cookies_out[cookie_name] = ""
        self.cookies_out[cookie_name]["path"] = "/"
        self.cookies_out[cookie_name]["expires"] = 0
        self.cookies_out[cookie_name]["max-age"] = 0

    # Application-specific methods.

    def set_user(self, username):

        """
        An application-specific method which sets the user information with
        'username' in the transaction. This affects subsequent calls to
        'get_user'.
        """

        self.user = username

# vim: tabstop=4 expandtab shiftwidth=4
